/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group (THG).                                         *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of eos52nc4. The full eos52nc4 copyright notice,      *
 * including terms governing use, modification, and redistribution, is       *
 * contained in the files COPYING and Copyright.html. COPYING can be         *
 * found at the root of the source code distribution tree; Copyright.html    *
 * can be found at the root level of an installed copy of the electronic     *
 * eos52nc4 document set and is linked from the top-level documents page.   *
 * It can also be found at                                                   *
 * http://www.hdfgroup.org/H4H5TOOLS/doc/Copyright.html. If you do not have  *
 * access to either file, you may request a copy from help@hdfgroup.org.     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "HE5_HdfEosDef.h"
#include "toy_util.h"
#include "toy_hdfeos5.h"
#include "toy_hdf5.h"
#include <stdio.h>


static int convert_onegrid(hid_t ctxid, const char *gridname)
{
  grid_t grid;
  int succeeded = 0;

  init_grid_t(&grid);
  do {
    if (read_hdfeos5_onegrid(ctxid, gridname, &grid) == -1) err1_break(3, "hdfeos5 read")
    if (write_hdf5_onegrid(ctxid, gridname, &grid) == -1) err1_break(3, "hdf5 write")

    succeeded = 1;
  } while (0);

  if (1) {
    free_grid_t(&grid);
  }
  return succeeded ? 0 : -1;
}

static int convert_oneswath(hid_t ctxid, const char *swathname)
{
  swath_t swath;
  int succeeded = 0;

  init_swath_t(&swath);
  do {
    if (read_hdfeos5_oneswath(ctxid, swathname, &swath) == -1) err1_break(3, "hdfeos5 read")
    if (write_hdf5_oneswath(ctxid, swathname, &swath) == -1) err1_break(3, "hdf5 write")

    succeeded = 1;
  } while (0);

  if (1) {
    free_swath_t(&swath);
  }
  return succeeded ? 0 : -1;
}

static int doit(const char *infname, const char *outfname)
{
  hid_t ctxid = -1;
  int succeeded = 0;

  do {
    size_t i;
    int failinside = 0;
    context_t *ctx = NULL;

    if ((ctxid = create_context()) == -1) earlyerr1_break(3, "create context")
    if ((ctx = get_context(ctxid)) == NULL) earlyerr1_break(3, "internal error?")

    if (open_hdfeos5(ctxid, infname) == -1) err1_break(3, "open hdfeos5 file")
    if (open_hdf5(ctxid, outfname) == -1) err1_break(3, "open hdf5 file")
    if (write_hdf5_globalattr(ctxid) == -1) err1_break(3, "write hdf5 global attributes")
    for (i = 0; i < ctx->gridlist.len; ++i) {
      if (convert_onegrid(ctxid, ctx->gridlist.s[i].name.s) == -1) err2_flagbreak(3, failinside, "grid", ctx->gridlist.s[i].name.s)
    }
    if (failinside) break;
    for (i = 0; i < ctx->swathlist.len; ++i) {
      if (convert_oneswath(ctxid, ctx->swathlist.s[i].name.s) == -1) err2_flagbreak(3, failinside, "swath", ctx->swathlist.s[i].name.s)
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  if (1) {
    close_hdf5(ctxid);
    close_hdfeos5(ctxid);
    close_context(ctxid);
  }
  return succeeded ? 0 : -1;
}

static void usage()
{
  printf("Usage: eos52nc4 <input_path> <output_path>\n");
}

int main(int argc, char **argv)
{
  if (argc != 3) {
    usage();
    return 1;
  }

  return doit(argv[1], argv[2]);
}

// vim:ts=8:sw=2:sts=2:cindent
