/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group (THG).                                         *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of eos52nc4. The full eos52nc4 copyright notice,      *
 * including terms governing use, modification, and redistribution, is       *
 * contained in the files COPYING and Copyright.html. COPYING can be         *
 * found at the root of the source code distribution tree; Copyright.html    *
 * can be found at the root level of an installed copy of the electronic     *
 * eos52nc4 document set and is linked from the top-level documents page.   *
 * It can also be found at                                                   *
 * http://www.hdfgroup.org/H4H5TOOLS/doc/Copyright.html. If you do not have  *
 * access to either file, you may request a copy from help@hdfgroup.org.     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define H5_USE_16_API 1
#include "HE5_HdfEosDef.h"

#define err_ret(s)  return _err(__FILE__, __LINE__, "", "", "")
#define err1_ret(s)  return _err(__FILE__, __LINE__, s, "", "")
#define err2_ret(s1, s2)  return _err(__FILE__, __LINE__, s1, s2, "")
#define err3_ret(s1, s2, s3)  return _err(__FILE__, __LINE__, s1, s2, s3)
#define suicide()  do { _err(__FILE__, __LINE__, "assertion failure", "committing suicide", ""); exit(1); } while(0)

static int _err(const char *fname, int line, const char *s1, const char *s2, const char *s3)
{
  fprintf(stderr, "failed at %s:%d - %s %s %s\n", fname, line, s1, s2, s3);
  return 1;
}

static herr_t HE5_GDdefdim_const(hid_t id, const char *name, hsize_t dim)
{
  return HE5_GDdefdim(id, (char *)name, dim);
}

static herr_t HE5_GDwriteattr_const(hid_t id, const char *name, hid_t ntype, const hsize_t count[], const void *data)
{
  return HE5_GDwriteattr(id, name, ntype, (hsize_t *)count, (void *)data);
}

static herr_t HE5_GDwritegrpattr_const(hid_t id, const char *name, hid_t ntype, const hsize_t count[], const void *data)
{
  return HE5_GDwritegrpattr(id, name, ntype, (hsize_t *)count, (void *)data);
}

static herr_t HE5_GDwritelocattr_const(hid_t id, const char *fieldname, const char *attrname, hid_t ntype, const hsize_t count[], const void *data)
{
  return HE5_GDwritelocattr(id, fieldname, attrname, ntype, (hsize_t *)count, (void *)data);
}

static herr_t HE5_GDdeffield_const(hid_t id, const char *name, const char *dimlist, const char *maxdimlist, hid_t type, int merge)
{
  return HE5_GDdeffield(id, name, (char *)dimlist, (char *)maxdimlist, type, merge);
}

static herr_t HE5_GDsetfillvalue_const(hid_t id, const char *name, hid_t ntype, const void *fillval)
{
  return HE5_GDsetfillvalue(id, name, ntype, (void *)fillval);
}

static herr_t HE5_GDwritefield_const(hid_t id, const char *name, const hssize_t start[], const hsize_t stride[], const hsize_t edge[], const void *data)
{
  return HE5_GDwritefield(id, name, start, stride, edge, (void *)data);
}


static herr_t HE5_SWdefdim_const(hid_t id, const char *name, hsize_t dim)
{
  return HE5_SWdefdim(id, (char *)name, dim);
}

static herr_t HE5_SWwritefield_const(hid_t id, const char *name, const hssize_t start[], const hsize_t stride[], const hsize_t edge[], const void *data)
{
  return HE5_SWwritefield(id, (char *)name, start, stride, edge, (void *)data);
}


static herr_t HE5_SWdefdimmap_const(hid_t id, const char *geoname, const char *dataname, hsize_t offset, hsize_t increment)
{
  return HE5_SWdefdimmap(id, (char *)geoname, (char *)dataname, offset, increment);
}

static herr_t HE5_SWwriteattr_const(hid_t id, const char *name, hid_t ntype, const hsize_t count[], const void *data)
{
  return HE5_SWwriteattr(id, name, ntype, (hsize_t *)count, (void *)data);
}

static herr_t HE5_SWwritegrpattr_const(hid_t id, const char *name, hid_t ntype, const hsize_t count[], const void *data)
{
  return HE5_SWwritegrpattr(id, name, ntype, (hsize_t *)count, (void *)data);
}

static herr_t HE5_SWwritelocattr_const(hid_t id, const char *fieldname, const char *attrname, hid_t ntype, const hsize_t count[], const void *data)
{
  return HE5_SWwritelocattr(id, fieldname, attrname, ntype, (hsize_t *)count, (void *)data);
}

static herr_t HE5_SWdefgeofield_const(hid_t id, const char *name, const char *dimlist, const char *maxdimlist, hid_t ntype, int merge)
{
  return HE5_SWdefgeofield(id, name, (char *)dimlist, (char *)maxdimlist, ntype, merge);
}

static herr_t HE5_SWdefdatafield_const(hid_t id, const char *name, const char *dimlist, const char *maxdimlist, hid_t ntype, int merge)
{
  return HE5_SWdefdatafield(id, name, (char *)dimlist, (char *)maxdimlist, ntype, merge);
}

static herr_t HE5_SWsetfillvalue_const(hid_t id, const char *name, hid_t ntype, const void *fillval)
{
  return HE5_SWsetfillvalue(id, (char *)name, ntype, (void *)fillval);
}

static herr_t HE5_EHwriteglbattr_const(hid_t id, const char *name, hid_t ntype, const hsize_t count[], const void *data)
{
  return HE5_EHwriteglbattr(id, name, ntype, (hsize_t *)count, (void *)data);
}


#define MAX_DIMS  5
#define MAX_FIELDS  20
#define MAX_GRIDS  15
#define MAX_SWATHS  15
#define MAX_DIMMAPS  5
#define MAX_BUFFER  5000
#define MAX_ATTRS  5

typedef struct {
  hid_t h5type;
  const char *str;
  const void *fillvalue;
} eostype_t;

static eostype_t eostypes[13];

static const char *fillvalue_STRING = "filler";

static void init_eostypes()
{
#define UNIVERSAL_FILLVALUE  62
  static const char fillvalue_CHAR = UNIVERSAL_FILLVALUE;
  static const unsigned char fillvalue_UCHAR = UNIVERSAL_FILLVALUE;
  static const short fillvalue_SHORT = UNIVERSAL_FILLVALUE;
  static const unsigned short fillvalue_USHORT = UNIVERSAL_FILLVALUE;
  static const int fillvalue_INT = UNIVERSAL_FILLVALUE;
  static const unsigned int fillvalue_UINT = UNIVERSAL_FILLVALUE;
  static const long fillvalue_LONG = UNIVERSAL_FILLVALUE;
  static const unsigned long fillvalue_ULONG = UNIVERSAL_FILLVALUE;
  static const long long fillvalue_LLONG = UNIVERSAL_FILLVALUE;
  static const unsigned long long fillvalue_ULLONG = UNIVERSAL_FILLVALUE;
  static const float fillvalue_FLOAT = UNIVERSAL_FILLVALUE;
  static const double fillvalue_DOUBLE = UNIVERSAL_FILLVALUE;
  /* NetCDF-4 does not allow LDOUBLE
  static const long double fillvalue_LDOUBLE = UNIVERSAL_FILLVALUE;
  */

  {
    int i = 0;
#define DEF_TYPE(he5, str, fill)  { eostype_t e = { he5, str, fill }; eostypes[i++] = e; }
    DEF_TYPE(H5T_NATIVE_CHAR, "char", &fillvalue_CHAR)
    DEF_TYPE(H5T_NATIVE_UCHAR, "uchar", &fillvalue_UCHAR)
    DEF_TYPE(H5T_NATIVE_SHORT, "short", &fillvalue_SHORT)
    DEF_TYPE(H5T_NATIVE_USHORT, "ushort", &fillvalue_USHORT)
    DEF_TYPE(H5T_NATIVE_INT, "int", &fillvalue_INT)
    DEF_TYPE(H5T_NATIVE_UINT, "uint", &fillvalue_UINT)
    DEF_TYPE(H5T_NATIVE_LONG, "long", &fillvalue_LONG)
    DEF_TYPE(H5T_NATIVE_ULONG, "ulong", &fillvalue_ULONG)
    DEF_TYPE(H5T_NATIVE_LLONG, "llong", &fillvalue_LLONG)
    DEF_TYPE(H5T_NATIVE_ULLONG, "ullong", &fillvalue_ULLONG)
    DEF_TYPE(H5T_NATIVE_FLOAT, "float", &fillvalue_FLOAT)
    DEF_TYPE(H5T_NATIVE_DOUBLE, "double", &fillvalue_DOUBLE)
    /* NetCDF-4 does not allow LDOUBLE
    DEF_TYPE(H5T_NATIVE_LDOUBLE, "ldouble", &fillvalue_LDOUBLE)
    */
    DEF_TYPE(H5T_C_S1, "string", fillvalue_STRING)
#undef DEF_TYPE
    if (i != 13) suicide();
  }
}

const char * get_stringliteral(int i)
{
  static const char *literals[] = {
    "literal0", "literal1", "literal2", "literal3", "literal4",
    "literal5", "literal6", "literal7", "literal8", "literal9",
  };
  return literals[i % (sizeof(literals) / sizeof(literals[0]))];
}

#define IMPLEMENT_TYPES() \
  HANDLE_PRIMITIVE_TYPE(H5T_NATIVE_CHAR, char) \
  HANDLE_PRIMITIVE_TYPE(H5T_NATIVE_UCHAR, unsigned char) \
  HANDLE_PRIMITIVE_TYPE(H5T_NATIVE_SHORT, short) \
  HANDLE_PRIMITIVE_TYPE(H5T_NATIVE_USHORT, unsigned short) \
  HANDLE_PRIMITIVE_TYPE(H5T_NATIVE_INT, int) \
  HANDLE_PRIMITIVE_TYPE(H5T_NATIVE_UINT, unsigned int) \
  HANDLE_PRIMITIVE_TYPE(H5T_NATIVE_LONG, long) \
  HANDLE_PRIMITIVE_TYPE(H5T_NATIVE_ULONG, unsigned long) \
  HANDLE_PRIMITIVE_TYPE(H5T_NATIVE_LLONG, long long) \
  HANDLE_PRIMITIVE_TYPE(H5T_NATIVE_ULLONG, unsigned long long) \
  HANDLE_PRIMITIVE_TYPE(H5T_NATIVE_FLOAT, float) \
  HANDLE_PRIMITIVE_TYPE(H5T_NATIVE_DOUBLE, double) \
  HANDLE_PRIMITIVE_TYPE(H5T_NATIVE_LDOUBLE, long double) \
  HANDLE_STRING_TYPE(H5T_C_S1)

static void fill_data(const hsize_t *dimsizes, hid_t datatype, int usefillvalue, void *data)
{
  hsize_t npoints = 1;
  size_t typesize;
  hsize_t i;

  typesize = H5Tget_size(datatype);
  for (i = 0; i < MAX_DIMS; ++i) {
    if (dimsizes[i] == 0) break;
    npoints *= dimsizes[i];
  }
  if (npoints * typesize > MAX_BUFFER) suicide();

  for (i = 0; i < npoints; ++i) {
    if (0) ;
#define HANDLE_PRIMITIVE_TYPE(eostype, ctype) \
    else if (H5Tequal(datatype, eostype) == TRUE) { ((ctype *)data)[i] = (ctype)(i % 2 == 0 ? i + 1 : (usefillvalue ? UNIVERSAL_FILLVALUE : i + 1)); }
#define HANDLE_STRING_TYPE(eostype) \
    else if (H5Tequal(datatype, eostype) == TRUE) { ((const char **)data)[i] = (const char *)(i % 2 == 0 ? get_stringliteral(i + 1) : (usefillvalue ? fillvalue_STRING : get_stringliteral(i + 1))); }
    IMPLEMENT_TYPES()
#undef HANDLE_PRIMITIVE_TYPE
#undef HANDLE_STRING_TYPE
    else suicide();
  }
}

static void fill_attr(hsize_t length, hid_t datatype, void *data)
{
  hsize_t i;
  size_t typesize;

  typesize = H5Tget_size(datatype);
  if (length * typesize > MAX_BUFFER) suicide();
  for (i = 0; i < length; ++i) {
    if (0) ;
#define HANDLE_PRIMITIVE_TYPE(eostype, ctype) \
    else if (H5Tequal(datatype, eostype) == TRUE) { ((ctype *)data)[i] = (ctype)(i + 1); }
#define HANDLE_STRING_TYPE(eostype) \
    else if (H5Tequal(datatype, eostype) == TRUE) { }
    IMPLEMENT_TYPES()
#undef HANDLE_PRIMITIVE_TYPE
#undef HANDLE_STRING_TYPE
    else suicide();
  }

  if (H5Tequal(datatype, H5T_C_S1) == TRUE) {
    static const char *attr = "attribute";
    for (i = 0; i < length; ++i)
      ((char *)data)[i] = attr[i % (sizeof(attr) / sizeof(attr[0]))];
    ((char *)data)[length] = '\0';
  }
}

typedef struct {
  const char *attrname;
  hsize_t length;
} grid_attr_t;

typedef struct {
  const char *fieldname;
  hid_t datatype;
  const char *dimnames;
  hsize_t dimsizes[MAX_DIMS];
  int use_fillvalue;
  const void *fillvalue;
  grid_attr_t locattrs[MAX_ATTRS];
} grid_field_t;

typedef struct {
  const char *dimname;
  hsize_t dimsize;
} grid_dim_t;

typedef struct {
  const char *gridname;
  long xdimsize;
  long ydimsize;
  int projcode;
  hid_t attrtype;
  grid_dim_t dims[MAX_DIMS];
  grid_field_t fields[MAX_FIELDS];
  grid_attr_t attrs[MAX_ATTRS];
  grid_attr_t grpattrs[MAX_ATTRS];
} eos_grid_t;

typedef struct {
  const char *filename;
  eos_grid_t grids[MAX_GRIDS];
  grid_attr_t glbattrs[MAX_ATTRS];
} grid_file_t;

typedef struct {
  int zonecode;
  int projcode;
  int spherecode;
  int origincode;
  double projparam[16];
  double upleft[2];
  double lowright[2];
} grid_proj_t;

static void fill_grid_proj(grid_proj_t *gridinfo, int projcode)
{
  int i;

  for (i = 0; i < 16; ++i)
    gridinfo->projparam[i] = 0;

  switch (projcode) {
    case HE5_GCTP_GEO:
      gridinfo->projcode = HE5_GCTP_GEO;
      gridinfo->zonecode = 0;
      gridinfo->spherecode = 0;
      gridinfo->origincode = 0;
      gridinfo->upleft[0] = HE5_EHconvAng(0.0, HE5_HDFE_DEG_DMS);
      gridinfo->upleft[1] = HE5_EHconvAng(30.0, HE5_HDFE_DEG_DMS);
      gridinfo->lowright[0] = HE5_EHconvAng(15.0, HE5_HDFE_DEG_DMS);
      gridinfo->lowright[1] = HE5_EHconvAng(20.0, HE5_HDFE_DEG_DMS);
      break;
    case HE5_GCTP_PS:
      gridinfo->projcode = HE5_GCTP_PS;
      gridinfo->zonecode = 0;
      gridinfo->spherecode = 3;
      gridinfo->origincode = HE5_HDFE_GD_LR;
      gridinfo->projparam[5] = 90000000.00;
      gridinfo->upleft[0] = 0;
      gridinfo->upleft[1] = 0;
      gridinfo->lowright[0] = 0;
      gridinfo->lowright[1] = 0;
      break;
    case HE5_GCTP_UTM:
      gridinfo->projcode = HE5_GCTP_UTM;
      gridinfo->zonecode = 40;
      gridinfo->spherecode = 0;
      gridinfo->origincode = 0;
      gridinfo->upleft[0] = 210584.50041;
      gridinfo->upleft[1] = 3322395.95445;
      gridinfo->lowright[0] = 813931.10959;
      gridinfo->lowright[1] = 2214162.53278;
      break;
    default:
      suicide();
  }
}

static int grid_generic(const grid_file_t *gridfile)
{
  hid_t fileid;
  int data[MAX_BUFFER];
  int i, j, k;

  if ((fileid = HE5_GDopen(gridfile->filename, H5F_ACC_TRUNC)) == -1) err2_ret("cannot create", gridfile->filename);
  for (i = 0; i < MAX_GRIDS; ++i) {
    const eos_grid_t *grid = &gridfile->grids[i];
    grid_proj_t proj;
    hid_t gdid;
    if (!grid->gridname) break;

    fill_grid_proj(&proj, grid->projcode);
    if ((gdid = HE5_GDcreate(fileid, grid->gridname, grid->xdimsize, grid->ydimsize, proj.upleft, proj.lowright)) == -1) err2_ret("cannot create", grid->gridname);
    if (HE5_GDdefproj(gdid, proj.projcode, proj.zonecode, proj.spherecode, proj.projparam) == -1) err1_ret("define projection");
    if (HE5_GDdeforigin(gdid, proj.origincode)) err1_ret("define origin");
    if (HE5_GDdefpixreg(gdid, HE5_HDFE_CENTER)) err1_ret("define pixreg");

    for (j = 0; j < MAX_DIMS; ++j) {
      const grid_dim_t *dim = &grid->dims[j];
      if (!dim->dimname) break;

      if (HE5_GDdefdim_const(gdid, dim->dimname, dim->dimsize) == -1) err2_ret("define dim", dim->dimname);
    }
    for (j = 0; j < MAX_ATTRS; ++j) {
      const grid_attr_t *attr = &grid->attrs[j];
      if (!attr->attrname) break;

      fill_attr(attr->length, grid->attrtype, data);
      if (HE5_GDwriteattr_const(gdid, attr->attrname, grid->attrtype, &attr->length, data) == -1) err2_ret("write attr", attr->attrname);
    }
    for (j = 0; j < MAX_ATTRS; ++j) {
      const grid_attr_t *attr = &grid->grpattrs[j];
      if (!attr->attrname) break;

      fill_attr(attr->length, grid->attrtype, data);
      if (HE5_GDwritegrpattr_const(gdid, attr->attrname, grid->attrtype, &attr->length, data) == -1) err2_ret("write grpattr", attr->attrname);
    }
    for (j = 0; j < MAX_FIELDS; ++j) {
      const grid_field_t *field = &grid->fields[j];
      if (!field->fieldname) break;

      if (field->use_fillvalue) {
	if (HE5_GDsetfillvalue_const(gdid, field->fieldname, field->datatype, field->fillvalue) == -1) err2_ret("set fillvalue", field->fieldname);
      }
      if (HE5_GDdeffield_const(gdid, field->fieldname, field->dimnames, NULL, field->datatype, HE5_HDFE_NOMERGE) == -1) err2_ret("define field", field->fieldname);

      for (k = 0; k < MAX_ATTRS; ++k) {
	const grid_attr_t *attr = &field->locattrs[k];
	if (!attr->attrname) break;

	fill_attr(attr->length, grid->attrtype, data);
	if (HE5_GDwritelocattr_const(gdid, field->fieldname, attr->attrname, grid->attrtype, &attr->length, data) == -1) err2_ret("write locattr", attr->attrname);
      }

      fill_data(field->dimsizes, field->datatype, field->use_fillvalue, data);
      if (HE5_GDwritefield_const(gdid, field->fieldname, NULL, NULL, NULL, data) == -1) err2_ret("write field", field->fieldname);
    }
    HE5_GDdetach(gdid);
  }
  for (i = 0; i < MAX_ATTRS; ++i) {
    const grid_attr_t *attr = &gridfile->glbattrs[i];
    if (!attr->attrname) break;

    fill_attr(attr->length, H5T_NATIVE_INT, data);
    if (HE5_EHwriteglbattr_const(fileid, attr->attrname, H5T_NATIVE_INT, &attr->length, data) == -1) err2_ret("write glbattr", attr->attrname);
  }
  HE5_GDclose(fileid);
  return 0;
}

typedef struct {
  const char *attrname;
  hsize_t length;
} swath_attr_t;

typedef struct {
  const char *fieldname;
  hid_t datatype;
  const char *dimnames;
  hsize_t dimsizes[MAX_DIMS];
  int use_fillvalue;
  const void *fillvalue;
  swath_attr_t locattrs[MAX_ATTRS];
} swath_field_t;

typedef struct {
  const char *dimname;
  hsize_t dimsize;
} swath_dim_t;

typedef struct {
  const char *geoname;
  const char *dataname;
  long offset;
  long increment;
} swath_dimmap_t;

typedef struct {
  const char *swathname;
  hid_t attrtype;
  swath_dim_t dims[MAX_DIMS];
  swath_dimmap_t dimmaps[MAX_DIMMAPS];
  swath_field_t geofields[MAX_FIELDS];
  swath_field_t datafields[MAX_FIELDS];
  swath_attr_t attrs[MAX_ATTRS];
  swath_attr_t grpattrs[MAX_ATTRS];
} eos_swath_t;

typedef struct {
  const char *filename;
  eos_swath_t swaths[MAX_SWATHS];
  swath_attr_t glbattrs[MAX_ATTRS];
} swath_file_t;

static hsize_t get_dimsizen(const eos_swath_t *swath, const char *dimname, size_t len)
{
  int i;

  for (i = 0; i < MAX_DIMS; ++i) {
    const swath_dim_t *dim = &swath->dims[i];
    if (!dim->dimname) break;

    if (strncmp(dim->dimname, dimname, len) == 0)
      return dim->dimsize;
  }
  suicide();
  return -1;
}

static hsize_t get_dimsize(const eos_swath_t *swath, const char *dimname)
{
  return get_dimsizen(swath, dimname, strlen(dimname));
}

static void validate_swath(const swath_file_t *swathfile)
{
  int i, j, k;

  for (i = 0; i < MAX_SWATHS; ++i) {
    const eos_swath_t *swath = &swathfile->swaths[i];
    if (!swath->swathname) break;

    /* dimension maps */
    for (j = 0; j < MAX_DIMS; ++j) {
      const swath_dimmap_t *dimmap = &swath->dimmaps[j];
      hsize_t geosize, datasize;
      if (!dimmap->geoname) break;

      geosize = get_dimsize(swath, dimmap->geoname);
      datasize = get_dimsize(swath, dimmap->dataname);

      /* The "offset" tells how far along a data dimension you must travel to find the */
      /* first point to have a corresponding entry along the geolocation dimension. */
      if (dimmap->offset > 0)
	datasize -= dimmap->offset;
      else if (dimmap->offset < 0)
	geosize -= -dimmap->offset;

      /* The "increment" tells how many points to travel along the data dimension before */
      /* the next point is found for which there is a corresponding entry along the geolocation dimension. */
      if (dimmap->increment > 1) {
	/* datasize = ( (geosize - 1) * increment, geosize * increment ] */
	if (!(((geosize - 1) * dimmap->increment < datasize) &&
	       (datasize <= geosize * dimmap->increment))) suicide();
      }
      else if (dimmap->increment < -1) {
	/* geosize = ( (datasize - 1) * increment, datasize * increment ] */
	if (!(((datasize - 1) * -dimmap->increment < geosize) &&
	       (geosize <= datasize * -dimmap->increment))) suicide();
      }
    }

    /* geo fields, data fields */
    for (j = 0; j < 2; ++j) {
      const swath_field_t *fields = j == 0 ? swath->geofields : swath->datafields;
      for (k = 0; k < MAX_FIELDS; ++k) {
	const swath_field_t *field = &fields[k];
	size_t dimnamelength, l, m, n = 0;
	if (!field->fieldname) break;

	dimnamelength = strlen(field->dimnames);
	for (l = m = 0; m <= dimnamelength; ++m) {
	  if ((m == dimnamelength && dimnamelength) || field->dimnames[m] == ',') {
	    hsize_t dimsize = get_dimsizen(swath, field->dimnames + l, m - l);
	    if (dimsize != field->dimsizes[n++]) suicide();

	    l = m + 1;
	  }
	}
      }
    }
  }
}

static int swath_generic(const swath_file_t *swathfile)
{
  hid_t fileid;
  int data[MAX_BUFFER];
  float geo[MAX_BUFFER];
  int i, j, k;

  validate_swath(swathfile);

  if ((fileid = HE5_SWopen(swathfile->filename, H5F_ACC_TRUNC)) == -1) err2_ret("cannot create", swathfile->filename);
  for (i = 0; i < MAX_SWATHS; ++i) {
    const eos_swath_t *swath = &swathfile->swaths[i];
    hid_t swid;
    if (!swath->swathname) break;

    if ((swid = HE5_SWcreate(fileid, swath->swathname)) == -1) err2_ret("cannot create", swath->swathname);

    for (j = 0; j < MAX_DIMS; ++j) {
      const swath_dim_t *dim = &swath->dims[j];
      if (!dim->dimname) break;

      if (HE5_SWdefdim_const(swid, dim->dimname, dim->dimsize) == -1) err2_ret("define dim", dim->dimname);
    }
    for (j = 0; j < MAX_DIMMAPS; ++j) {
      const swath_dimmap_t *dimmap = &swath->dimmaps[j];
      if (!dimmap->geoname) break;

      if (HE5_SWdefdimmap_const(swid, dimmap->geoname, dimmap->dataname, dimmap->offset, dimmap->increment) == -1) err3_ret("define dimmap", dimmap->geoname, dimmap->dataname);
    }
    for (j = 0; j < MAX_ATTRS; ++j) {
      const swath_attr_t *attr = &swath->attrs[j];
      if (!attr->attrname) break;

      fill_attr(attr->length, swath->attrtype, data);
      if (HE5_SWwriteattr_const(swid, attr->attrname, swath->attrtype, &attr->length, data) == -1) err2_ret("write attr", attr->attrname);
    }
    for (j = 0; j < MAX_ATTRS; ++j) {
      const swath_attr_t *attr = &swath->grpattrs[j];
      if (!attr->attrname) break;

      fill_attr(attr->length, swath->attrtype, data);
      if (HE5_SWwritegrpattr_const(swid, attr->attrname, swath->attrtype, &attr->length, data) == -1) err2_ret("write grpattr", attr->attrname);
    }
    for (j = 0; j < MAX_FIELDS; ++j) {
      const swath_field_t *field = &swath->geofields[j];
      if (!field->fieldname) break;

      if (field->use_fillvalue) {
	if (HE5_SWsetfillvalue_const(swid, field->fieldname, field->datatype, field->fillvalue) == -1) err2_ret("set fillvalue", field->fieldname);
      }
      if (HE5_SWdefgeofield_const(swid, field->fieldname, field->dimnames, NULL, field->datatype, HE5_HDFE_NOMERGE) == -1) err2_ret("define field", field->fieldname);

      for (k = 0; k < MAX_ATTRS; ++k) {
	const swath_attr_t *attr = &field->locattrs[k];
	if (!attr->attrname) break;

	fill_attr(attr->length, swath->attrtype, data);
	if (HE5_SWwritelocattr_const(swid, field->fieldname, attr->attrname, swath->attrtype, &attr->length, data) == -1) err2_ret("write locattr", attr->attrname);
      }

      fill_data(field->dimsizes, field->datatype, field->use_fillvalue, geo);
      if (HE5_SWwritefield_const(swid, field->fieldname, NULL, NULL, NULL, geo) == -1) err2_ret("write", field->fieldname);
    }
    for (j = 0; j < MAX_FIELDS; ++j) {
      const swath_field_t *field = &swath->datafields[j];
      if (!field->fieldname) break;

      if (field->use_fillvalue) {
	if (HE5_SWsetfillvalue_const(swid, field->fieldname, field->datatype, field->fillvalue) == -1) err2_ret("set fillvalue", field->fieldname);
      }
      if (HE5_SWdefdatafield_const(swid, field->fieldname, field->dimnames, NULL, field->datatype, HE5_HDFE_NOMERGE) == -1) err2_ret("define field", field->fieldname);

      for (k = 0; k < MAX_ATTRS; ++k) {
	const swath_attr_t *attr = &field->locattrs[k];
	if (!attr->attrname) break;

	fill_attr(attr->length, swath->attrtype, data);
	if (HE5_SWwritelocattr_const(swid, field->fieldname, attr->attrname, swath->attrtype, &attr->length, data) == -1) err2_ret("write locattr", attr->attrname);
      }

      fill_data(field->dimsizes, field->datatype, field->use_fillvalue, data);
      if (HE5_SWwritefield_const(swid, field->fieldname, NULL, NULL, NULL, data) == -1) err2_ret("write", field->fieldname);
    }
    HE5_SWdetach(swid);
  }
  for (i = 0; i < MAX_ATTRS; ++i) {
    const swath_attr_t *attr = &swathfile->glbattrs[i];
    if (!attr->attrname) break;

    fill_attr(attr->length, H5T_NATIVE_INT, data);
    if (HE5_EHwriteglbattr_const(fileid, attr->attrname, H5T_NATIVE_INT, &attr->length, data) == -1) err2_ret("write glbattr", attr->attrname);
  }
  HE5_SWclose(fileid);
  return 0;
}

int main(int argc, char **argv)
{
  init_eostypes();

#define ASSERT_SUCCESS(x)  \
  if (x) { fprintf(stderr, "failed to generate %s\n", eos.filename); return 1; } \
  else { printf("generated %s\n", eos.filename); }

#pragma region GENERIC {{{
#pragma region Type {{{
#pragma region Grid {{{
  {
    int i;
    eos_grid_t grid =
	{ "", 5, 7, HE5_GCTP_GEO, 0,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", 0, "XDim,YDim", { 5, 7 }, 0, 0 },
	    { "field1_fill", 0, "XDim,YDim", { 5, 7 }, 1, 0 },
	    { 0, 0, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "attr1", 3 },
	  }
	};
    grid_file_t eos;
    memset(&eos, 0, sizeof(eos));
    eos.filename = "generic_type_grid.he5";

    for (i = 0; i < sizeof(eostypes) / sizeof(eostypes[0]); ++i) {
      eos.grids[i] = grid;
      eos.grids[i].gridname = eostypes[i].str;
      eos.grids[i].attrtype = eostypes[i].h5type;
      eos.grids[i].fields[0].datatype = eostypes[i].h5type;
      eos.grids[i].fields[1].datatype = eostypes[i].h5type;
      eos.grids[i].fields[1].fillvalue = eostypes[i].fillvalue;
    }
    eos.glbattrs[0].attrname = "glbattr1";
    eos.glbattrs[0].length = 3;
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}

#pragma region Swath {{{
  {
    int i;
    eos_swath_t swath =
	{ "", 0,
	  {
	    { "dim0", 3 },
	    { "dim1", 5 },
	    { "dim2", 7 },
	  },
	  {
	    { 0, 0, 0, 0 },
	  },
	  {
	    { "Time", H5T_NATIVE_FLOAT, "dim0", { 3 }, 0, 0 },
	    { "Longitude", H5T_NATIVE_FLOAT, "dim1,dim2", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "dim1,dim2", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", 0, "dim0", { 3 }, 0, 0 },
	    { "field1_fill", 0, "dim0", { 3 }, 1, 0 },
	    { "field2", 0, "dim1,dim2", { 5, 7 }, 0, 0 },
	    { "field2_fill", 0, "dim1,dim2", { 5, 7 }, 1, 0 },
	    { 0, 0, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "attr1", 3 },
	  }
	};
    swath_file_t eos;
    memset(&eos, 0, sizeof(eos));
    eos.filename = "generic_type_swath.he5";

    for (i = 0; i < sizeof(eostypes) / sizeof(eostypes[0]); ++i) {
      eos.swaths[i] = swath;
      eos.swaths[i].swathname = eostypes[i].str;
      eos.swaths[i].attrtype = eostypes[i].h5type;
      eos.swaths[i].datafields[0].datatype = eostypes[i].h5type;
      eos.swaths[i].datafields[1].datatype = eostypes[i].h5type;
      eos.swaths[i].datafields[1].fillvalue = eostypes[i].fillvalue;
      eos.swaths[i].datafields[2].datatype = eostypes[i].h5type;
      eos.swaths[i].datafields[3].datatype = eostypes[i].h5type;
      eos.swaths[i].datafields[3].fillvalue = eostypes[i].fillvalue;
    }
    eos.glbattrs[0].attrname = "glbattr1";
    eos.glbattrs[0].length = 3;
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}

#pragma region Shared name {{{
#pragma region Grid {{{
  {
    /* It turns out that this is not a valid file because
     * every field should refer to at least XDim and YDim.
     */
    grid_file_t eos =
    { "grid_sharedname.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_GEO, H5T_NATIVE_INT,
	  {
	    { "dim0", 9 },
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "XDim,YDim", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "dim0", { 9 }, 0, 0 },
	    { "XDim", H5T_NATIVE_INT, "XDim", { 5 }, 0, 0 },
	    { "YDim", H5T_NATIVE_INT, "YDim", { 7 }, 0, 0 },
	    { "dim0", H5T_NATIVE_INT, "dim0", { 9 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	},
	{ "grid2", 5, 7, HE5_GCTP_UTM, H5T_NATIVE_INT,
	  {
	    { "dim0", 9 },
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "XDim,YDim", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "dim0", { 9 }, 0, 0 },
	    { "XDim", H5T_NATIVE_INT, "XDim", { 5 }, 0, 0 },
	    { "YDim", H5T_NATIVE_INT, "YDim", { 7 }, 0, 0 },
	    { "dim0", H5T_NATIVE_INT, "dim0", { 9 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}

#pragma region Swath {{{
  {
    swath_file_t eos =
    { "swath_sharedname.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "dim0", 5 },
	    { "dim1", 7 },
	    { "dim2", 9 },
	  },
	  {
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "dim0,dim1", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "dim0,dim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "dim0,dim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "dim0,dim1", { 5, 7 }, 0, 0 },
	    { "dim2", H5T_NATIVE_INT, "dim2", { 9 } , 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}

#pragma region Bad name {{{
  /* HDF-EOS2 allows slashes in names, but HDF-EOS5 does not */
#pragma region Grid {{{
  if (0) {
    grid_file_t eos =
    { "grid_badname.he5",
      {
	{ "grid1*asterisk", 5, 7, HE5_GCTP_GEO, H5T_NATIVE_INT,
	  {
	    { "dim0*asterisk", 3 },
	    { 0, 0 },
	  },
	  {
	    { "field1*asterisk", H5T_NATIVE_INT, "XDim,YDim", { 5, 7 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { "field2*asterisk", H5T_NATIVE_INT, "dim0*asterisk", { 3 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { "field3*asterisk", H5T_NATIVE_INT, "XDim,YDim,dim0*asterisk", { 5, 7, 3 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "attr1*asterisk", 3 },
	    { 0, 0 },
	  },
	  {
	    { "grpattr1*asterisk", 3 },
	    { 0, 0 },
	  }
	},
	{ "grid2", 5, 7, HE5_GCTP_UTM, H5T_NATIVE_INT,
	  {
	    { "dim0*asterisk", 3 },
	    { 0, 0 },
	  },
	  {
	    { "field1*asterisk", H5T_NATIVE_INT, "XDim,YDim", { 5, 7 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { "field2*asterisk", H5T_NATIVE_INT, "dim0*asterisk", { 3 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { "field3*asterisk", H5T_NATIVE_INT, "XDim,YDim,dim0*asterisk", { 5, 7, 3 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "attr1*asterisk", 3 },
	    { 0, 0 },
	  },
	  {
	    { "grpattr1*asterisk", 3 },
	    { 0, 0 },
	  }
	}
      },
      {
	{ "glbattr1*asterisk", 3 },
	{ 0, 0 },
      }
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}

#pragma region Swath {{{
  if (0) {
    swath_file_t eos =
    { "swath_badname.he5",
      {
	{ "swath1*asterisk", H5T_NATIVE_INT,
	  {
	    { "dim0*asterisk", 5 },
	    { "dim1*asterisk", 7 },
	  },
	  {
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "dim0*asterisk,dim1*asterisk", { 5, 7 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { "Latitude", H5T_NATIVE_FLOAT, "dim0*asterisk,dim1*asterisk", { 5, 7 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "dim0*asterisk,dim1*asterisk", { 5, 7 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { "field2", H5T_NATIVE_INT, "dim0*asterisk,dim1*asterisk", { 5, 7 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "attr1*asterisk", 3 },
	    { 0, 0 },
	  },
	  {
	    { "grpattr1*asterisk", 3 },
	    { 0, 0 },
	  }
	},
	{ "swath2*asterisk", H5T_NATIVE_INT,
	  {
	    { "geodim0_underscore", 4 },
	    { "geodim1_underscore", 6 },
	    { "datadim0_underscore", 5 },
	    { "datadim1_underscore", 7 },
	  },
	  {
	    /* When dimension maps are used, slash cannot be used because this is the delimiter.
	     * I can parse more smartly, but isn't it overkill? */
	    { "geodim0_underscore", "datadim0_underscore", 1, 1 },
	    { "geodim1_underscore", "datadim1_underscore", 1, 1 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0_underscore,geodim1_underscore", { 4, 6 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0_underscore,geodim1_underscore", { 4, 6 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0_underscore,datadim1_underscore", { 5, 7 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { "field2", H5T_NATIVE_INT, "datadim0_underscore,datadim1_underscore", { 5, 7 }, 0, 0, { { "locattr1*asterisk", 3 }, { 0, 0 } } },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "attr1*asterisk", 3 },
	    { 0, 0 },
	  },
	  {
	    { "grpattr1*asterisk", 3 },
	    { 0, 0 },
	  }
	}
      },
      {
	{ "glbattr1*asterisk", 3 },
	{ 0, 0 },
      }
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}
#pragma region Attribute {{{
  {
    grid_file_t eos =
    { "grid_attr.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_GEO, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "XDim,YDim", { 5, 7 }, 0, 0, { { "locattr1", 3 }, { 0, 0 } } },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "attr1", 3 },
	    { 0, 0 },
	  },
	  {
	    { "grpattr1", 3 },
	    { 0, 0 },
	  }
	},
      },
      {
	{ "glbattr1", 3 },
	{ 0, 0 },
      }
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }

  {
    swath_file_t eos =
    { "swath_attr.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "dim0", 5 },
	    { "dim1", 7 },
	  },
	  {
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "dim0,dim1", { 5, 7 }, 0, 0, { { "locattr1", 3 }, { 0, 0 } } },
	    { "Latitude", H5T_NATIVE_FLOAT, "dim0,dim1", { 5, 7 }, 0, 0, { { "locattr1", 3 }, { 0, 0 } } },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "dim0,dim1", { 5, 7 }, 0, 0, { { "locattr1", 3 }, { 0, 0 } } },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "attr1", 3 },
	    { 0, 0 },
	  },
	  {
	    { "grpattr1", 3 },
	    { 0, 0 },
	  }
	},
      },
      {
	{ "glbattr1", 3 },
	{ 0, 0 },
      }
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}

#pragma region GRID {{{
#pragma region GEO {{{
#pragma region V(XDim, YDim) GCTP_GEO {{{
  {
    grid_file_t eos =
    { "grid_geo_xy.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_GEO, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "XDim,YDim", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma region V(YDim, XDim) GCTP_GEO {{{
  {
    grid_file_t eos =
    { "grid_geo_yx.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_GEO, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "YDim,XDim", { 7, 5 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma region V(Time, XDim, YDim) GCTP_GEO {{{
  {
    grid_file_t eos =
    { "grid_geo_timexy.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_GEO, H5T_NATIVE_INT,
	  {
	    { "Time", 3 },
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "Time,XDim,YDim", { 3, 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma region V(Time, YDim, XDim) GCTP_GEO {{{
  {
    grid_file_t eos =
    { "grid_geo_timeyx.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_GEO, H5T_NATIVE_INT,
	  {
	    { "Time", 3 },
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "Time,YDim,XDim", { 3, 7, 5 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}

#pragma region UTM {{{
#pragma region V(XDim, YDim) GCTP_UTM (2-D) {{{
  {
    grid_file_t eos =
    { "grid_utm_xy.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_UTM, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "XDim,YDim", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma region V(YDim, XDim) GCTP_UTM (2-D) {{{
  {
    grid_file_t eos =
    { "grid_utm_yx.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_UTM, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "YDim,XDim", { 7, 5 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma region V(Time, XDim, YDim) GCTP_UTM (2-D) {{{
  {
    grid_file_t eos =
    { "grid_utm_timexy.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_UTM, H5T_NATIVE_INT,
	  {
	    { "Time", 3 },
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "Time,XDim,YDim", { 3, 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma region V(Time, YDim, XDim) GCTP_UTM (2-D) {{{
  {
    grid_file_t eos =
    { "grid_utm_timeyx.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_UTM, H5T_NATIVE_INT,
	  {
	    { "Time", 3 },
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "Time,YDim,XDim", { 3, 7, 5 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}

#pragma region Multiple variables {{{
#pragma region GEO {{{
#pragma region V(XDim, YDim) W(XDim, YDim) GCTP_GEO {{{
  {
    grid_file_t eos =
    { "grid_geo_xyxy.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_GEO, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "XDim,YDim", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "XDim,YDim", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma region V(YDim, XDim) W(YDim, XDim) GCTP_GEO {{{
  {
    grid_file_t eos =
    { "grid_geo_yxyx.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_GEO, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "YDim,XDim", { 7, 5 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "YDim,XDim", { 7, 5 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma region V(XDim, YDim) W(YDim, XDim) GCTP_GEO (conflict) {{{
  {
    grid_file_t eos =
    { "grid_geo_xyyx.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_GEO, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "XDim,YDim", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "YDim,XDim", { 7, 5 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma region no fields GCTP_GEO (cannot guess) {{{
  {
    grid_file_t eos =
    { "grid_geo_nil.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_GEO, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}

#pragma region UTM {{{
#pragma region V(XDim, YDim) W(XDim, YDim) GCTP_UTM (2-D) {{{
  {
    grid_file_t eos =
    { "grid_utm_xyxy.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_UTM, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "XDim,YDim", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "XDim,YDim", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma region V(YDim, XDim) W(YDim, XDim) GCTP_UTM (2-D) {{{
  {
    grid_file_t eos =
    { "grid_utm_yxyx.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_UTM, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "YDim,XDim", { 7, 5 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "YDim,XDim", { 7, 5 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma region V(XDim, YDim) W(YDim, XDim) GCTP_UTM (conflict) {{{
  {
    grid_file_t eos =
    { "grid_utm_xyyx.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_UTM, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "XDim,YDim", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "YDim,XDim", { 7, 5 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma region no fields GCTP_UTM (cannot guess) {{{
  {
    grid_file_t eos =
    { "grid_utm_nil.he5",
      {
	{ "grid1", 5, 7, HE5_GCTP_UTM, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}
#pragma endregion }}}

#pragma region Multiple Grids {{{
  {
    grid_file_t eos =
    { "grid_multgrids.he5",
      {
	{ "grid1", 1, 3, HE5_GCTP_GEO, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "XDim,YDim", { 1, 3 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "XDim,YDim", { 1, 3 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "Attribute1", 1 },
	    { "Attribute2", 3 },
	  }
	},
	{ "grid2", 5, 7, HE5_GCTP_UTM, H5T_NATIVE_INT,
	  {
	    { 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "YDim,XDim", { 7, 5 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "YDim,XDim", { 7, 5 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "Attribute1", 5 },
	    { "Attribute2", 7 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(grid_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}

#pragma region SWATH {{{
#pragma region dimension map {{{
#pragma region without dimmap {{{
  {
    swath_file_t eos =
    { "swath_basic.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "dim0", 5 },
	    { "dim1", 7 },
	  },
	  {
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "dim0,dim1", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "dim0,dim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "dim0,dim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "dim0,dim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}

#pragma region without scaling {{{
#pragma region dimmap (offset: 0, increment: 1) {{{
  {
    swath_file_t eos =
    { "swath_0101.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 5 },
	    { "geodim1", 7 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", 0, 1 },
	    { "geodim1", "datadim1", 0, 1 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma region dimmap (offset: 1, increment: 1) {{{
  {
    swath_file_t eos =
    { "swath_1111.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 4 },
	    { "geodim1", 6 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", 1, 1 },
	    { "geodim1", "datadim1", 1, 1 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 4, 6 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 4, 6 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma region dimmap (offset: 2, increment: 1) {{{
  {
    swath_file_t eos =
    { "swath_2121.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 3 },
	    { "geodim1", 5 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", 2, 1 },
	    { "geodim1", "datadim1", 2, 1 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 3, 5 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 3, 5 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma region dimmap (offset: -1, increment: 1) {{{
  {
    swath_file_t eos =
    { "swath_n11n11.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 6 },
	    { "geodim1", 8 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", -1, 1 },
	    { "geodim1", "datadim1", -1, 1 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 6, 8 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 6, 8 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma region dimmap (offset: -1, increment: 1) geodim is not bigger than datadim {{{
  {
    swath_file_t eos =
    { "swath_n11n11_tight.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 5 },
	    { "geodim1", 7 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", -1, 1 },
	    { "geodim1", "datadim1", -1, 1 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma region dimmap (offset: -2, increment: 1) {{{
  {
    swath_file_t eos =
    { "swath_n21n21.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 7 },
	    { "geodim1", 9 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", -2, 1 },
	    { "geodim1", "datadim1", -2, 1 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 7, 9 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 7, 9 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}

#pragma region scaling (increment > 1) {{{
#pragma region dimmap (offset: 0, increment: 2) {{{
  {
    swath_file_t eos =
    { "swath_0202.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 3 },
	    { "geodim1", 4 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", 0, 2 },
	    { "geodim1", "datadim1", 0, 2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 3, 4 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 3, 4 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma region dimmap (offset: 1, increment: 2) {{{
  {
    swath_file_t eos =
    { "swath_1212.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 2 },
	    { "geodim1", 3 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", 1, 2 },
	    { "geodim1", "datadim1", 1, 2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 2, 3 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 2, 3 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma region dimmap (offset: 2, increment: 2) {{{
  {
    swath_file_t eos =
    { "swath_2222.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 2 },
	    { "geodim1", 3 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", 2, 2 },
	    { "geodim1", "datadim1", 2, 2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 2, 3 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 2, 3 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma region dimmap (offset: -1, increment: 2) {{{
  {
    swath_file_t eos =
    { "swath_n12n12.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 4 },
	    { "geodim1", 5 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", -1, 2 },
	    { "geodim1", "datadim1", -1, 2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 4, 5 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 4, 5 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma region dimmap (offset: -2, increment: 2) {{{
  {
    swath_file_t eos =
    { "swath_n22n22.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 5 },
	    { "geodim1", 6 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", -2, 2 },
	    { "geodim1", "datadim1", -2, 2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 6 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 6 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}

#pragma region scaling (increment < -1) {{{
#pragma region dimmap (offset: 0, increment: -2) {{{
  {
    swath_file_t eos =
    { "swath_0n20n2.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 9 },
	    { "geodim1", 13 }, 
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", 0, -2 },
	    { "geodim1", "datadim1", 0, -2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 9, 13 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 9, 13 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma region dimmap (offset: 1, increment: -2) {{{
  {
    swath_file_t eos =
    { "swath_1n21n2.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 7 },
	    { "geodim1", 11 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", 1, -2 },
	    { "geodim1", "datadim1", 1, -2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 7, 11 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 7, 11 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma region dimmap (offset: 2, increment: -2) {{{
  {
    swath_file_t eos =
    { "swath_2n22n2.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 6 },
	    { "geodim1", 10 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", 2, -2 },
	    { "geodim1", "datadim1", 2, -2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 6, 10 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 6, 10 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma region dimmap (offset: -1, increment: -2) {{{
  {
    swath_file_t eos =
    { "swath_n1n2n1n2.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 10 },
	    { "geodim1", 14 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", -1, -2 },
	    { "geodim1", "datadim1", -1, -2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 10, 14 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 10, 14 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma region dimmap (offset: -2, increment: -2) {{{
  {
    swath_file_t eos =
    { "swath_n2n2n2n2.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "geodim0", 11 },
	    { "geodim1", 15 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", -2, -2 },
	    { "geodim1", "datadim1", -2, -2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 11, 15 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 11, 15 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "field2", H5T_NATIVE_INT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}

#pragma region multiple geofields for a datafield {{{
#pragma region one data field without dimmap {{{
#pragma region qual, datadim0, datadim1 {{{
  {
    swath_file_t eos =
    { "swath_qdd.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "qualdim", 3 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "qualdim,datadim0,datadim1", { 3, 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}

#pragma region datadim0, qual, datadim1 {{{
  {
    swath_file_t eos =
    { "swath_dqd.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "qualdim", 3 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,qualdim,datadim1", { 5, 3, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}

#pragma region datadim0, datadim1, qual {{{
  {
    swath_file_t eos =
    { "swath_ddq.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "qualdim", 3 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "datadim0,datadim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1,qualdim", { 5, 7, 3 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}

#pragma region one data field with a dimmap {{{
#pragma region qual, geodim0, geodim1 {{{
  {
    swath_file_t eos =
    { "swath_qgg.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "qualdim", 3 },
	    { "geodim0", 5 },
	    { "geodim1", 7 },
	    { "datadim0", 10 },
	    { "datadim1", 14 },
	  },
	  {
  	    { "geodim0", "datadim0", 0, 2 },
	    { "geodim1", "datadim1", 0, 2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "qualdim,datadim0,datadim1", { 3, 10, 14 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}

#pragma region geodim0, qual, geodim1 {{{
  {
    swath_file_t eos =
    { "swath_gqg.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "qualdim", 3 },
	    { "geodim0", 5 },
	    { "geodim1", 7 },
	    { "datadim0", 10 },
	    { "datadim1", 14 },
	  },
	  {
  	    { "geodim0", "datadim0", 0, 2 },
	    { "geodim1", "datadim1", 0, 2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,qualdim,datadim1", { 10, 3, 14 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}

#pragma region geodim0, geodim1, qual {{{
  {
    swath_file_t eos =
    { "swath_ggq.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "qualdim", 3 },
	    { "geodim0", 5 },
	    { "geodim1", 7 },
	    { "datadim0", 10 },
	    { "datadim1", 14 },
	  },
	  {
  	    { "geodim0", "datadim0", 0, 2 },
	    { "geodim1", "datadim1", 0, 2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field1", H5T_NATIVE_INT, "datadim0,datadim1,qualdim", { 10, 14, 3 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}

#pragma region Qual, Longitude, Latitude {{{
  {
    swath_file_t eos =
    { "swath_tll.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "qualdim", 3 },
	    { "geodim0", 5 },
	    { "geodim1", 7 },
	    { "datadim0", 5 },
	    { "datadim1", 7 },
	  },
	  {
	    { "geodim0", "datadim0", 0, 1 },
	    { "geodim1", "datadim1", 0, 1 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field01", H5T_NATIVE_INT, "qualdim", { 3 }, 0, 0 },
	    { "field02", H5T_NATIVE_INT, "geodim0", { 5 }, 0, 0 },
	    { "field03", H5T_NATIVE_INT, "geodim1", { 7 }, 0, 0 },
	    { "field04", H5T_NATIVE_INT, "datadim0", { 5 }, 0, 0 },
	    { "field05", H5T_NATIVE_INT, "datadim1", { 7 }, 0, 0 },
	    { "field06", H5T_NATIVE_INT, "qualdim,geodim0", { 3, 5 }, 0, 0 },
	    { "field07", H5T_NATIVE_INT, "qualdim,geodim1", { 3, 7 }, 0, 0 },
	    { "field08", H5T_NATIVE_INT, "qualdim,datadim0", { 3, 5 }, 0, 0 },
	    { "field09", H5T_NATIVE_INT, "qualdim,datadim1", { 3, 7 }, 0, 0 },
	    { "field10", H5T_NATIVE_INT, "geodim0,qualdim", { 5, 3 }, 0, 0 },
	    { "field11", H5T_NATIVE_INT, "geodim1,qualdim", { 7, 3 }, 0, 0 },
	    { "field12", H5T_NATIVE_INT, "datadim0,qualdim", { 5, 3 }, 0, 0 },
	    { "field13", H5T_NATIVE_INT, "datadim1,qualdim", { 7, 3 }, 0, 0 },
	    { "field14", H5T_NATIVE_INT, "qualdim,geodim0,geodim1", { 3, 5, 7 }, 0, 0 },
	    { "field15", H5T_NATIVE_INT, "qualdim,datadim0,datadim1", { 3, 5, 7 }, 0, 0 },
	    { "field16", H5T_NATIVE_INT, "geodim0,qualdim,geodim1", { 5, 3, 7 }, 0, 0 },
	    { "field17", H5T_NATIVE_INT, "datadim0,qualdim,datadim1", { 5, 3, 7 }, 0, 0 },
	    { "field18", H5T_NATIVE_INT, "geodim0,geodim1,qualdim", { 5, 7, 3 }, 0, 0 },
	    { "field19", H5T_NATIVE_INT, "datadim0,datadim1,qualdim", { 5, 7, 3 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}

#pragma region Qual, Longitude, Latitude with scaling {{{
  {
    swath_file_t eos =
    { "swath_tll02.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "qualdim", 3 },
	    { "geodim0", 5 },
	    { "geodim1", 7 },
	    { "datadim0", 10 },
	    { "datadim1", 14 },
	  },
	  {
	    { "geodim0", "datadim0", 0, 2 },
	    { "geodim1", "datadim1", 0, 2 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field01", H5T_NATIVE_INT, "qualdim", { 3 }, 0, 0 },
	    { "field02", H5T_NATIVE_INT, "geodim0", { 5 }, 0, 0 },
	    { "field03", H5T_NATIVE_INT, "geodim1", { 7 }, 0, 0 },
	    { "field04", H5T_NATIVE_INT, "datadim0", { 10 }, 0, 0 },
	    { "field05", H5T_NATIVE_INT, "datadim1", { 14 }, 0, 0 },
	    { "field06", H5T_NATIVE_INT, "qualdim,geodim0", { 3, 5 }, 0, 0 },
	    { "field07", H5T_NATIVE_INT, "qualdim,geodim1", { 3, 7 }, 0, 0 },
	    { "field08", H5T_NATIVE_INT, "qualdim,datadim0", { 3, 10 }, 0, 0 },
	    { "field09", H5T_NATIVE_INT, "qualdim,datadim1", { 3, 14 }, 0, 0 },
	    { "field10", H5T_NATIVE_INT, "geodim0,qualdim", { 5, 3 }, 0, 0 },
	    { "field11", H5T_NATIVE_INT, "geodim1,qualdim", { 7, 3 }, 0, 0 },
	    { "field12", H5T_NATIVE_INT, "datadim0,qualdim", { 10, 3 }, 0, 0 },
	    { "field13", H5T_NATIVE_INT, "datadim1,qualdim", { 14, 3 }, 0, 0 },
	    { "field14", H5T_NATIVE_INT, "qualdim,geodim0,geodim1", { 3, 5, 7 }, 0, 0 },
	    { "field15", H5T_NATIVE_INT, "qualdim,datadim0,datadim1", { 3, 10, 14 }, 0, 0 },
	    { "field16", H5T_NATIVE_INT, "geodim0,qualdim,geodim1", { 5, 3, 7 }, 0, 0 },
	    { "field17", H5T_NATIVE_INT, "datadim0,qualdim,datadim1", { 10, 3, 14 }, 0, 0 },
	    { "field18", H5T_NATIVE_INT, "geodim0,geodim1,qualdim", { 5, 7, 3 }, 0, 0 },
	    { "field19", H5T_NATIVE_INT, "datadim0,datadim1,qualdim", { 10, 14, 3 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { 0, 0 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}
#pragma endregion }}}

#pragma region Multiple Swaths {{{
  {
    swath_file_t eos =
    { "swath_multswaths.he5",
      {
	{ "swath1", H5T_NATIVE_INT,
	  {
	    { "qualdim", 2 },
	    { "geodim0", 4 },
	    { "geodim1", 6 },
	    { "datadim0", 4 },
	    { "datadim1", 6 },
	  },
	  {
	    { "geodim0", "datadim0", 0, 1 },
	    { "geodim1", "datadim1", 0, 1 },
	    { 0, 0 },
	  },
	  {
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 4, 6 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 4, 6 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field01", H5T_NATIVE_INT, "qualdim", { 2 }, 0, 0 },
	    { "field02", H5T_NATIVE_INT, "datadim0,datadim1", { 4, 6 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "Attribute1", 2 },
	    { "Attribute2", 4 },
	  }
	},
	{ "swath2", H5T_NATIVE_INT,
	  {
	    { "qualdim", 3 },
	    { "geodim0", 5 },
	    { "geodim1", 7 },
	  },
	  {
	    { 0, 0 },
	  },
	  {
	    { "Qual", H5T_NATIVE_FLOAT, "qualdim", { 3, }, 0, 0 },
	    { "Longitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { "Latitude", H5T_NATIVE_FLOAT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "field01", H5T_NATIVE_INT, "qualdim", { 3 }, 0, 0 },
	    { "field02", H5T_NATIVE_INT, "geodim0,geodim1", { 5, 7 }, 0, 0 },
	    { 0, H5T_NATIVE_INT, 0, { 0, }, 0, 0 },
	  },
	  {
	    { "Attribute1", 3 },
	    { "Attribute2", 5 },
	  }
	}
      },
    };
    ASSERT_SUCCESS(swath_generic(&eos))
  }
#pragma endregion }}}
#pragma endregion }}}

  return 0;
}

/* vim:set ts=8 sw=2 sts=2 cindent fdm=marker: */
